
#plot_specden <- function(x){
#    par(mfrow=c(1,2))
#    fields::image.plot( Re(x) )
#    fields::image.plot( Im(x) )
#    par(mfrow=c(1,1))
#}

##############################################
##### penalties, derivatives, hessian
##############################################

pen_logdet_cross_spec <- function(covparms, effrange, nwrap = 2 ){
    # compute the cross spectral densities and their cholesky decomps
    #cross_spec <- compute_cross_spec( covparms, effrange, nwrap )
    cross_spec <- matern_cross_spectra( covparms, effrange )
    chol_spec <- cholesky_array( cross_spec )
    # return the negative sum_logdet penalty
    return( sum_logdet( chol_spec ) )
}

pen_traceinv_cross_spec <- function(covparms, effrange, nwrap = 2 ){
    # compute the cross spectral densities and their cholesky decomps
    #cross_spec <- compute_cross_spec( covparms, effrange, nwrap )
    cross_spec <- matern_cross_spectra( covparms, effrange )
    marg_spec <- matern_marg_spectra2( covparms, effrange )
    inv_cross_spec <- inverse_array( cross_spec )
    M <- array_mat_mult( inv_cross_spec, marg_spec )
    # return the negative sum_logdet penalty
    return( array_trace( M ) )
}

dpen_traceinv_cross_spec <- function(covparms, effrange, nwrap = 2 ){
    # compute the cross spectral densities and their cholesky decomps
    #cross_spec <- compute_cross_spec( covparms, effrange, nwrap )
    nparms <- length(covparms)
    cross_spec <- matern_cross_spectra( covparms, effrange )
    marg_spec <- matern_marg_spectra2( covparms, effrange )
    d_cross_spec <- d_matern_cross_spectra( covparms, effrange )
    d_marg_spec <- d_matern_marg_spectra2( covparms, effrange )
    inv_cross_spec <- inverse_array( cross_spec )
    dpn <- rep(NA, nparms)
    for(j in 1:nparms){
        M1 <- array_mat_mult( inv_cross_spec, d_cross_spec[,,,j] )
        M1 <- array_mat_mult( M1, inv_cross_spec )
        M1 <- array_mat_mult( M1, marg_spec )
        M2 <- array_mat_mult( inv_cross_spec, d_marg_spec[,,,j] )
        dpn[j] <- array_trace( -M1 + M2 )
    }
    # return the negative sum_logdet penalty
    return(dpn)
}

ddpen_traceinv_cross_spec <- function(covparms, effrange, nwrap = 2 ){
    # compute the cross spectral densities and their cholesky decomps
    #cross_spec <- compute_cross_spec( covparms, effrange, nwrap )
    nparms <- length(covparms)
    cross_spec <- matern_cross_spectra( covparms, effrange )
    d_cross_spec <- d_matern_cross_spectra( covparms, effrange )
    dd_cross_spec <- dd_matern_cross_spectra( covparms, effrange )
    marg_spec <- matern_marg_spectra2( covparms, effrange )
    d_marg_spec <- d_matern_marg_spectra2( covparms, effrange )
    dd_marg_spec <- dd_matern_marg_spectra2( covparms, effrange )
    inv_cross_spec <- inverse_array( cross_spec )
    ddpn <- matrix(NA, nparms, nparms)
    for(j1 in 1:nparms){
        for(j2 in 1:nparms){

            A1 <- array_mat_mult( inv_cross_spec, d_cross_spec[,,,j1] )
            A2 <- array_mat_mult( inv_cross_spec, d_cross_spec[,,,j2] )

            A12 <- array_mat_mult( inv_cross_spec, dd_cross_spec[,,,j1,j2] )

            B <- array_mat_mult( inv_cross_spec, marg_spec )

            B1 <- array_mat_mult( inv_cross_spec, d_marg_spec[,,,j1] )
            B2 <- array_mat_mult( inv_cross_spec, d_marg_spec[,,,j2] )

            B12 <- array_mat_mult( inv_cross_spec, dd_marg_spec[,,,j1,j2] )


            A1_A2 <- array_mat_mult( A1, A2 )
            A1_A2_B <- array_mat_mult( A1_A2, B )

            A12_B <- array_mat_mult( A12, B )

            A2_A1 <- array_mat_mult( A2, A1 )
            A2_A1_B <- array_mat_mult( A2_A1, B )

            A1_B2 <- array_mat_mult( A1, B2 )

            A2_B1 <- array_mat_mult( A2, B1 )

            # B12 <- B12

            ddpn[j1,j2] <- array_trace( A1_A2_B + A2_A1_B - A12_B - A1_B2 - A2_B1 + B12 )
        }
    }
    return(ddpn)
}

pen_logvar_marg_spec <- function(covparms, effrange, nwrap = 2 ){
    #marg_spec <- compute_marg_spec( covparms, effrange, nwrap )
    marg_spec <- matern_marg_spectra( covparms, effrange )
    return( sum_logvar( marg_spec ) )
}

dpen_logdet_cross_spec <- function(covparms, effrange, nwrap = 2 ){
    # compute the cross spectral densities and their cholesky decomps

    nparms <- length(covparms)
    #cross_spec <- compute_cross_spec( covparms, effrange, nwrap )
    cross_spec <- matern_cross_spectra( covparms, effrange )
    chol_spec <- cholesky_array( cross_spec )
    d_cross_spec <- d_matern_cross_spectra( covparms, effrange )
    # compute Li_Sj_Lit
    LSL <- array(NA, dim(d_cross_spec))
    for(j in 1:nparms){ LSL[,,,j] <- Li_Sj_Lit( chol_spec, d_cross_spec[,,,j] ) }
    # compute derivative of logdets using LSL
    dpn <- rep(NA, nparms )
    for(j in 1:nparms){ dpn[j] <- d_sum_logdet( LSL[,,,j] ) }
    return(dpn)
}

dpen_logvar_marg_spec <- function(covparms, effrange, nwrap = 2 ){
    nparms <- length(covparms)
    #marg_spec <- compute_marg_spec( covparms, effrange, nwrap )
    marg_spec <- matern_marg_spectra( covparms, effrange )
    #d_marg_spec <- compute_d_marg_spec( covparms, effrange, nwrap )
    d_marg_spec <- d_matern_marg_spectra( covparms, effrange )
    dpn <- rep(NA,nparms)
    for(j in 1:nparms){ dpn[j] <- d_sum_logvar( marg_spec, d_marg_spec[,,j] ) }
    return(dpn)
}

ddpen_logdet_cross_spec <- function(covparms, effrange, nwrap = 2 ){
    # compute the cross spectral densities and their cholesky decomps

    nparms <- length(covparms)
    #cross_spec <- compute_cross_spec( covparms, effrange, nwrap )
    cross_spec <- matern_cross_spectra( covparms, effrange )
    chol_spec <- cholesky_array( cross_spec )
    # compute the derivative of the cross spectral densities
    #d_cross_spec <- compute_d_cross_spec( covparms, effrange, nwrap )
    d_cross_spec <- d_matern_cross_spectra( covparms, effrange )
    dd_cross_spec <- dd_matern_cross_spectra( covparms, effrange )
    # compute Li_Sj_Lit
    LSL <- array(NA, dim(d_cross_spec))
    for(j in 1:nparms){ LSL[,,,j] <- Li_Sj_Lit( chol_spec, d_cross_spec[,,,j] ) }

    ddpn <- array(0, c(nparms,nparms) )
    for(j1 in 1:nparms){
        for(j2 in 1:nparms){
            LSjkL <- Li_Sj_Lit( chol_spec, dd_cross_spec[,,,j1,j2] )
            ddpn[j1,j2] <- dd_sum_logdet( LSL[,,,j1], LSL[,,,j2], LSjkL )
        }
    }
    return(ddpn)
}

ddpen_logvar_marg_spec <- function(covparms, effrange, nwrap = 2 ){
    nparms <- length(covparms)
    #marg_spec <- compute_marg_spec( covparms, effrange, nwrap )
    marg_spec <- matern_marg_spectra( covparms, effrange )
    d_marg_spec <- d_matern_marg_spectra( covparms, effrange )
    dd_marg_spec <- dd_matern_marg_spectra( covparms, effrange )
    ddpn <- matrix(0, nparms, nparms )
    for(j1 in 1:nparms){ for(j2 in 1:nparms){
        ddpn[j1,j2] <-
            dd_sum_logvar(marg_spec,d_marg_spec[,,j1],d_marg_spec[,,j2],dd_marg_spec[,,j1,j2] )
    }}
    return(ddpn)
}


approx_max_range <- function(locs){
    locs0 <- locs[ sample(1:nrow(locs),400), ]
    return( max( fields::rdist(locs0) ) )
}

##############################################
##### cross spectral density and derivatives
##############################################

compute_cross_spec <- function(covparms, effrange, nwrap = 2){

    # matrix of covparms and number of components
    parm_mat <- multi_parms_mat( covparms )
    ncomp <- nrow( parm_mat$var )

    # number of gridpoints in the integration
    # also deserves some scrutiny
    ngrid <- 100

    # initialize array of ncomp x ncomp cross spectral density matrices
    cross_spec <- array( NA, c(ngrid,ngrid,ncomp,ncomp) )

    for( j1 in 1:ncomp){ for(j2 in 1:ncomp){
        # use a nugget of zero because we want to evaluate whether
        # the underlying matern process is valid, not allow an
        # invalid matern process to be "saved" by adding a nugget
        matern_parms <- c(parm_mat$var[j1,j2], parm_mat$range[j1,j2], parm_mat$smoothness[j1,j2], 0)
        cov_arr <- wrap_cov( matern_parms, ngrid, effrange, nwrap )
        cross_spec[,,j1,j2] <- fft(cov_arr)
    }}
    return(cross_spec)
}


compute_marg_spec <- function(covparms, effrange, nwrap = 2){

    # matrix of covparms and number of components
    parm_mat <- multi_parms_mat( covparms )
    ncomp <- nrow( parm_mat$var )

    # number of gridpoints in the integration
    # also deserves some scrutiny
    ngrid <- 100

    # initialize array of marginal spectral densities
    marg_spec <- array( NA, c(ngrid,ngrid,ncomp) )

    for( j1 in 1:ncomp){
        # use a nugget of zero because we want to evaluate whether
        # the underlying matern process is valid, not allow an
        # invalid matern process to be "saved" by adding a nugget
        matern_parms <- c(parm_mat$var[j1,j1], parm_mat$range[j1,j1], parm_mat$smoothness[j1,j1], 0)
        cov_arr <- wrap_cov( matern_parms, ngrid, effrange, nwrap )
        marg_spec[,,j1] <- Re( fft(cov_arr) )
    }
    return(marg_spec)
}


compute_d_cross_spec <- function( covparms, effrange, nwrap = 2 ){

    # matrix of covparms and number of components
    parm_mat <- multi_parms_mat( covparms )
    ncomp <- nrow( parm_mat$var )

    # number of gridpoints in the integration
    # also deserves some scrutiny
    ngrid <- 100
    d_cross_spec <- array(0, c( ngrid, ngrid, ncomp, ncomp, length(covparms) ) )

    # take fft of derivatives of covariances with respect to each parameter
    # for every pair of components
    # this one only has to loop over every unique pair, i.e. j2 <= j1
    for( j1 in 1:ncomp){ for(j2 in 1:j1){
        matern_parms <- c(parm_mat$var[j1,j2], parm_mat$range[j1,j2], parm_mat$smoothness[j1,j2], 0)
        d_cov_arr <- d_wrap_cov( matern_parms, ngrid, effrange, nwrap )
        # figure how to assign to the right index
        index <- multi_matern_parm_index( ncomp, j1, j2 )
        for( k in 1:3){
            i <- index[[k]]
            d_cross_spec[,,j1,j2,i] <- fft(d_cov_arr[,,k])
            if(j1 != j2){ d_cross_spec[,,j2,j1,i] <- Conj(d_cross_spec[,,j1,j2,i]) }
        }
    }}
    return(d_cross_spec)
}



compute_d_marg_spec <- function( covparms, effrange, nwrap = 2 ){

    # matrix of covparms and number of components
    parm_mat <- multi_parms_mat( covparms )
    ncomp <- nrow( parm_mat$var )

    # number of gridpoints in the integration
    # also deserves some scrutiny
    ngrid <- 100
    d_marg_spec <- array(0, c( ngrid, ngrid, ncomp, length(covparms) ) )

    # take fft of derivatives of covariances with respect to each parameter
    # for every component
    for( j1 in 1:ncomp){
        matern_parms <- c(parm_mat$var[j1,j1], parm_mat$range[j1,j1], parm_mat$smoothness[j1,j1], 0)
        d_cov_arr <- d_wrap_cov( matern_parms, ngrid, effrange, nwrap )
        # figure how to assign to the right index
        index <- multi_matern_parm_index( ncomp, j1, j1 )
        for( k in 1:3){
            i <- index[[k]]
            d_marg_spec[,,j1,i] <- Re( fft(d_cov_arr[,,k]) )
        }
    }
    return(d_marg_spec)
}



##############################################
##### lower-level penalties
##############################################

sum_logdet <- function( cholcs ){
    ss <- 0
    dimcs <- dim(cholcs)
    ncomp <- dimcs[ length(dimcs) ]

    for(j in 1:ncomp ){
        ss <- ss + 2*sum( log( Re(cholcs[,j,j]) ) )
    }
    return(ss)
}

d_sum_logdet <- function( LSLj ){
    return( array_trace( LSLj ) )
}

dd_sum_logdet <- function( LSjL, LSkL, LSjkL ){
    LSjLLSkL <- array_mat_mult( LSjL, LSkL )
    return( array_trace( LSjkL - LSjLLSkL ) )
}



sum_logvar <- function( marg_spec ){
    return( sum(log(marg_spec)) )
}

d_sum_logvar <- function( marg_spec, d_marg_spec_j ){
    return( sum( d_marg_spec_j/marg_spec ) )
}

dd_sum_logvar <- function( marg_spec, d_marg_spec_j, d_marg_spec_k, dd_marg_spec_jk ){
    # not the true second derivative for j = k, but perhaps close
    return( sum( dd_marg_spec_jk/marg_spec - d_marg_spec_j*d_marg_spec_k/marg_spec^2 ) )
}


##############################################
##### linear algebra functions
##############################################

array_trace <- function(A){
    # array of matrices A
    dimA <- dim(A)
    ncomp <- dimA[ length(dimA) ]
    ss <- 0
    for(k in 1:ncomp){ ss <- ss + sum(Re(A[,k,k])) }
    return(ss)
}
    


forward_solve <- function( mat, vec ){
    n <- length(vec)
    sol <- rep(NA, n)
    sol[1] <- vec[1]/mat[1,1]

    for(j in 2:n){
        sol[j] <- ( vec[j] - sol[1:(j-1)] %*% mat[j,1:(j-1)] )/mat[j,j]
    }
    return(c(sol))
}


array_mat_mult <- function( A, B ){

    # arrays of matrices A and B
    # get the dimensions
    dimA <- dim(A)
    ncomp <- dimA[ length(dimA) ]
    
    # initialize array of products
    P <- array(0, dimA )

    # do the multiplication
    for(j1 in 1:ncomp){ for(j2 in 1:ncomp){
        temp <- P[,j1,j2]
        for(j3 in 1:ncomp){
            temp <- temp + A[,j1,j3]*B[,j3,j2]
        }
        P[,j1,j2] <- temp
    }}
    return(P)
}


forward_solve_array <- function( L, B ){

    # array of lower triangular matrices L, array of right hand side matrices B

    # get the dimensions
    dimL <- dim(L)
    ncomp <- dimL[ length(dimL) ]
    
    # array of solutions
    sol <- array(0, dimL )

    # do each solve separately. sol and B should always have last index set to k
    for(k in 1:ncomp){
        
        sol[,1,k] <- B[,1,k]/L[,1,1]

        for(j in 2:ncomp){
            temp <- B[,j,k]
            for( el in 1:(j-1) ){
                temp <- temp - sol[,el,k]*L[,j,el]
            }
            sol[,j,k] <- temp/L[,j,j]
        }
    }

    return(sol)
}

inverse_array <- function(cs){
    inv <- array(NA, dim(cs))
    for(j in 1:dim(cs)[1]){
        inv[j,,] <- solve( cs[j,,] )
    }
    return(inv)
}

cholesky_array <- function( cs ){

    # get the dimensions
    dimcs <- dim(cs)
    ncomp <- dimcs[ length(dimcs) ]
    
    # initialize cholesky array
    cholcs <- array(0, dimcs) 

    # (1,1) entry
    arr <- cs[,1,1]
    bad <- abs( Im(arr) ) > 1e-8 | Re(arr) < 0
    fill <- sqrt(cs[,1,1])
    fill[bad] <- 0
    cholcs[,1,1] <- fill
    
    if( ncomp == 1 ){ return( cholcs ) }

    # (2,1) entry
    cholcs[,2,1] <- cs[,2,1]/cholcs[,1,1]
    # (2,2) entry
    arr <- cs[,2,2] - cholcs[,2,1]*Conj( cholcs[,2,1] )
    bad <- abs( Im(arr) ) > 1e-8 | Re(arr) < 0
    fill <- sqrt(arr)
    fill[bad] <- 0
    cholcs[,2,2] <- fill

    if( ncomp == 2 ){ return( cholcs ) }

    # loop over rows
    for(j in 3:ncomp){

        # do a forward solve to get the first j-1 columns of row j # a^j = L^(j-1) ell^j
        # get entry (j,1)
        cholcs[,j,1] <- cs[,1,j]/cholcs[,1,1]

        # entries (j,2) through (j,j-1)
        for( k in 2:(j-1) ){

            temp <- cs[,j,k]

            # subtract off inner product, doing each term in a loop
            for( el in 1:(k-1) ){
                temp <- temp - cholcs[,j,el]*cholcs[,k,el]
            }

            # divide by the diagonal
            cholcs[,j,k] <- temp/cholcs[,k,k]
        }

        # get entry (j,j)
        temp <- cs[,j,j]
        for( k in 1:(j-1)){
            temp <- temp - cholcs[,j,k]*Conj( cholcs[,j,k] )
        }
        #cholcs[,,j,j] <- sqrt( temp )

        arr <- temp
        bad <- abs( Im(arr) ) > 1e-8 | Re(arr) < 0
        fill <- sqrt(arr)
        fill[bad] <- 0
        cholcs[,j,j] <- fill
    }
    
    return(cholcs)
}


Li_Sj_Lit <- function( cholcs, d_cross_specden_j ){

    # get the dimensions
    dimcs <- dim(cholcs)
    ncomp <- dimcs[ length(dimcs) ]
    
    A <- forward_solve_array( cholcs, d_cross_specden_j )
    A <- aperm( A, c(1,3,2) )
    A <- forward_solve_array( cholcs, A )
    A <- aperm( A, c(1,3,2) )
    return(A)

}








##############################################
##### parameter index operations
##############################################

# get ncomp from nparms
ncomp_from_nparms <- function( nparms ){
    # length(covparms) is 3*ncomp*(ncomp+1)/2 + ncomp
    ncomp <- NA
    for(j in 1:20){
        if( nparms == 4*j*(j+1)/2 ){
            ncomp <- j
        }
    }
    if( is.na(ncomp) ){ stop("could not determine number of components") }
    return(ncomp)
}


ncomp_from_nlogparms <- function( nlogparms ){
    # length(covparms) is 3*ncomp*(ncomp+1)/2 + ncomp
    ncomp <- NA
    for(j in 1:20){
        if(  nlogparms == (2 + 4*(j*(j+1)/2)) ){
            ncomp <- j
        }
    }
    if( is.na(ncomp) ){ stop("could not determine number of components") }
    return(ncomp)
}

ncomp_from_nlogparms_pars <- function( nlogparms ){    
    ncomp <- NA    
    for(j in 1:20){    
        if( nlogparms == (j + 1 + 2*(j*(j+1)/2)) ){    
            ncomp <- j    
        }    
    }    
    if( is.na(ncomp) ){ stop("could not determine number of components") }    
    return(ncomp)    
}    

multi_parms_mat <- function( covparms ){
    ncomp <- ncomp_from_nparms( length(covparms) )

    vmat <- matrix(NA, ncomp, ncomp)
    vst <- 0
    rmat <- matrix(NA, ncomp, ncomp)
    rst <- ncomp*(ncomp+1)/2
    smat <- matrix(NA, ncomp, ncomp)
    sst <- 2*rst
    nmat <- matrix(NA, ncomp, ncomp)
    nst <- 3*rst
    cnt <- 0
    for(j1 in 1:ncomp){
        for(j2 in 1:j1){
            k <- (j1-1)*j1/2 + j2         
            vmat[j1,j2] <- covparms[vst + k]
            vmat[j2,j1] <- covparms[vst + k]
            rmat[j1,j2] <- covparms[rst + k]
            rmat[j2,j1] <- covparms[rst + k]
            smat[j1,j2] <- covparms[sst + k]
            smat[j2,j1] <- covparms[sst + k]
            nmat[j1,j2] <- covparms[nst + k]
            nmat[j2,j1] <- covparms[nst + k]
        }
    }
    return( list( variance=vmat, range=rmat, smoothness=smat, nugget = nmat ) )
}

multi_parms_vec <- function( parm_mats ){

    ncomp <- nrow( parm_mats$variance )
    nmultiparms <- 3*ncomp*(ncomp+1)/2
    inds_mat <- matrix( 1:(ncomp^2), ncomp, ncomp )
    mat_inds <- 1
    for(j in 2:ncomp){ mat_inds <- c(mat_inds, inds_mat[j,1:j]) }
    
    parm_vec <- rep(NA, nmultiparms)
    inds <- 1:(ncomp*(ncomp+1)/2)

    cur_ind <- 0
    parm_vec[ cur_ind + inds ] <- parm_mats$variance[ mat_inds ]

    cur_ind <- ncomp*(ncomp+1)/2
    parm_vec[ cur_ind + inds ] <- parm_mats$range[ mat_inds ]

    cur_ind <- 2*ncomp*(ncomp+1)/2
    parm_vec[ cur_ind + inds ] <- parm_mats$smoothness[ mat_inds ]

    return( parm_vec )
}


multi_matern_parm_index <- function( ncomp, j1, j2 ){

    # assumes j1 >= j2, fix if not
    if( j2 > j1 ){
        tmp <- j2
        j2 <- j1
        j1 <- tmp
    }

    v1 <- 0
    r1 <- ncomp*(ncomp+1)/2
    s1 <- 2*r1
    n1 <- 3*r1

    k <- (j1-1)*j1/2 + j2
    index <- list()
    index$variance   <- v1 + k
    index$range      <- r1 + k
    index$smoothness <- s1 + k
    index$nugget     <- n1 + k
    
    return(index)
}


##############################################
##### wrapping functions
##############################################

wrap_cov <- function( matern_parms, ngrid, effrange, nwrap = 1 ){
    xgrid <- as.matrix( expand.grid( (0:(ngrid-1))*effrange/ngrid, (0:(ngrid-1))*effrange/ngrid) )
    cov_vec <- array(0, nrow(xgrid) )
    for( j1 in (-nwrap):(nwrap-1) ){
        for( j2 in (-nwrap):(nwrap-1) ){
            xx <- xgrid
            xx[,1] <- xgrid[,1] + j1*effrange
            xx[,2] <- xgrid[,2] + j2*effrange
            cov_vec <- cov_vec + c(matern_isotropic2( matern_parms, matrix(0,1,2), xx ))
        }
    }
    cov_arr <- array( cov_vec, c(ngrid,ngrid) )
    return(cov_arr)
}

    
d_wrap_cov <- function( matern_parms, ngrid, effrange, nwrap = 1 ){
    nparms <- length(matern_parms)
    xgrid <- as.matrix( expand.grid( (0:(ngrid-1))*effrange/ngrid, (0:(ngrid-1))*effrange/ngrid) )
    d_cov_vec <- array(0, c(nrow(xgrid),nparms))
    for( j1 in (-nwrap):(nwrap-1) ){
        for( j2 in (-nwrap):(nwrap-1) ){
            xx <- xgrid
            xx[,1] <- xgrid[,1] + j1*effrange
            xx[,2] <- xgrid[,2] + j2*effrange
            d_cov_vec <- d_cov_vec + c(d_matern_isotropic2( matern_parms, matrix(0,1,2), xx ))
        }
    }
    d_cov_arr <- array( NA, c(ngrid,ngrid,nparms) )
    for(j in 1:nparms){
        d_cov_arr[ , , j ] <- array( d_cov_vec[,j], c(ngrid,ngrid) )
    }
    return(d_cov_arr)
}


##############################################
##### extract components into list
##############################################


to.list <- function(y, X, locs){
	
    ncomp <- length(unique(locs[,ncol(locs)]))  
    ly <- list()
    lX <- list()
    llocs <- list()
    for(i in 1:ncomp){
	  inds <- which(locs[, ncol(locs)] == i) 
	  ly[[i]] <- y[inds]
	  lX[[i]] <- X[inds,]
	  llocs[[i]] <- locs[inds, 1:(ncol(locs)-1)]
    }
	
    return(list(ly = ly, lX  = lX, llocs = llocs))

}

##############################################
##### link functions
##############################################

f <- function(theta){


    ncomp <- NA
    for(j in 1:20){
        if( length(theta) == j*(j-1)/2 ){
            ncomp <- j
        }
    }

    L <- matrix(0,ncomp,ncomp)
    L[upper.tri(L, diag=FALSE )] <- theta
    diag(L) <- 1
    L <- t(L)
    norms <- apply(L, 1, function(x) norm(x, type = "2"))
    L <- L / norms
    M <- L%*%t(L)
}


finv <- function(M){
	
    ncomp <- nrow(M)
    tol <- 1e-8
    eigenvalues <- eigen(M, only.values = TRUE)$values

    for ( i in 1: ncomp ) {
        if ( abs( eigenvalues[i] ) < tol ) {
            eigenvalues[i] <- 0
        }
    }    
    if ( any( eigenvalues < 0 ) ) {
        theta <- rep(NA, ncomp*(ncomp-1)/2)
    }else{
    	L <- t(chol(M))
    	L <- L * kronecker(t(rep(1, ncomp)), rep(1, ncomp)/diag(L))
    	theta <- t(L)[upper.tri(L)]
    }
}


